open Ast
open Sast

(* Map from variable name to its type *)
module NameMap = Map.Make(String)

(* Set of object names that are the same type *)
module ObjectTypeSet = Set.Make(String)

(* module ObjectFieldSet = Set.Make((String * Sast.data_type)) *)
module ObjectFieldSet = Set.Make(String)

module StringSet = Set.Make(String)

type environment = data_type NameMap.t
type id = string
type substitutions = (id * data_type) list

let rec infer_all (ast: Ast.program) =
  let sast_types = infer_types ast in
  let sast_objects, objs = infer_objs sast_types in
  (sast_objects, objs, obj_to_id_hash)

and logging = false

and log (msg: string) =
  if logging then print_endline msg else ()

and print_field_hash () =
  let _ = log "<field_hash>" in
  let _ = Hashtbl.iter (fun k v -> log (k ^ "->" ^ (string_of_data_type v))) field_hash in
  let _ = log "</field_hash>\n" in ()

and print_set (set: ObjectTypeSet.t) =
  let _ = log "<set>" in
  let list = ObjectTypeSet.elements set in
  let _ = List.iter log list in
  let _ = log "</set>" in ()

and print_set_list (set_list: ObjectTypeSet.t list) =
  let _ = log "\n<set list>" in
  let _ = List.iter print_set set_list in
  let _ = log "</set list>\n" in ()

and print_obj_to_fields_hash () =
  let _ = log "\n<obj_to_fields_hash>" in
  let _ = Hashtbl.iter (fun k v -> let _ = log (k ^ "->") in let _ = List.iter log v in ()) obj_to_fields_hash in
  let _ = log "</obj_to_fields_hash>\n" in ()

and print_obj_to_id_hash () =
  let _ = log "\n<obj_to_id_hash>" in
  let _ = Hashtbl.iter (fun k v -> let _ = log (k ^ "->" ^ (string_of_int v)) in ()) obj_to_id_hash in
  let _ = log "</obj_to_id_hash>\n" in ()

and print_id_to_fields_hash () =
  let _ = log "\n<id_to_fields_hash>" in
  let _ = Hashtbl.iter (fun k v -> let _ = log ((string_of_int k) ^ "->") in let _ = List.iter log v in ()) id_to_fields_hash in
  let _ = log "</id_to_fields_hash>\n" in ()

and print_id_to_fields_typified_hash () =
  let _ = log "\n<id_to_fields_typified_hash>" in
  let _ = Hashtbl.iter (fun k v -> let _ = log ((string_of_int k) ^ "->") in List.iter (fun x -> log ("(" ^ (fst x) ^ "," ^ string_of_data_type(snd x) ^ ")")) v) id_to_fields_typified_hash in
  let _ = log "</id_to_fields_typified_hash>\n" in ()


(*
infer_types: Driver for type inference - given a program's AST, turns it into an sast
Params:
  ast: AST for a Pseudo program (list of fdecls)
Returns:
  fdecls: SAST for a Pseudo program (list of afdecls)
*)
and infer_types (ast: Ast.program) =
  let _ = List.iter add_func_to_map ast
  and _ = if not (Hashtbl.mem func_decl_hash "main") then
    raise (failwith ("main function not found"))
  (* Begin with main() function *)

  and _ = infer_types_function (Hashtbl.find func_decl_hash "main") []
  and _  = add_afunc_decls()
  and _ = print_field_hash()
  and fdecls = order_fdecls ast in
  fdecls


and order_fdecls (ast: Ast.program) =
  let _ = Hashtbl.add func_order_hash "main" 0
  and fdecl_tups = List.fold_left get_name_to_order [] ast
  in let sorted_tups = List.sort compare fdecl_tups
  in let fdecls = List.map drop_order sorted_tups in fdecls

and get_name_to_order acc fdecl =
    if Hashtbl.mem afunc_decl_hash fdecl.fname then
      (let tup = (Hashtbl.find afunc_decl_hash fdecl.fname, Hashtbl.find
                func_order_hash fdecl.fname)
      in tup :: acc)
    else acc

(* drop the order from the tuple *)
and drop_order x =  match x with
    (fdecl, _) -> fdecl

(* compare by second key *)
and compare x y =
  snd x - snd y
(*
get_afunc_decl: Given a list of key (function name) value (function type) pairs, outputs an afdecl list
Looks through the global hash tables to collect information about functions
Params:
  k: Function Name
  v: Function Type
Returns:
  l: List of afdecls representing the sast
*)

and add_afunc_decl (k: string) (b: astmt list) =
  let ret = Hashtbl.find func_type_hash k
  and aforms = Hashtbl.find func_formals_hash k in
  let new_fdecl = {
    afname = k;
    aformals = aforms;
    abody = b;
    return = ret
  } in Hashtbl.add afunc_decl_hash new_fdecl.afname new_fdecl


(*
collect_afunc_decls: Driver to collect a list of afdecls from the global hash maps
Returns:
  l: List of afdecls representing the sast
*)
and add_afunc_decls ()=
  Hashtbl.iter add_afunc_decl func_body_hash

(*
add_func_to_map: Adds a funcdecl to the global declaration hash
This hash map is isued to get parameters when a call is detected, mandating that that function be analyzed
Returns:
  unit: Never actually used
*)
and add_func_to_map (fd: func_decl) =
  Hashtbl.add func_decl_hash fd.fname fd

(*
gen_new_placeholder: Uses characters for placeholder types in inference for a single statement
Returns:
  c1: Character used as a placeholder
*)
and gen_new_placeholder () =
  let c1 = !type_variable in
  incr type_variable; T(Char.escaped (Char.chr c1))

and gen_new_index (): int =
  let n1 = !index in
  let _ = incr index in n1


(* Placeholder character *)
and type_variable = ref (Char.code 'a')

and index = ref 1

(* Hash Table for function declarations *)
and func_decl_hash = Hashtbl.create 10
(* Hash Table for function types after inference *)
and func_type_hash = Hashtbl.create 10
(* Hash Table for annotated formal parameters [eg add(a, b) -> int add(int a, int b)] *)
and func_formals_hash = Hashtbl.create 10
(* Hash Table for function bodies, in the form of list of annotated statements *)
and func_body_hash = Hashtbl.create 10
(* Hash Table for function order *)
and func_order_hash = Hashtbl.create 10
(* Hash Table for function decl *)
and afunc_decl_hash = Hashtbl.create 10
(* Hash Table for object fields *)
and field_hash = Hashtbl.create 30

and fun_vis_rec = Hashtbl.create 10
and func_env_hash = Hashtbl.create 10


and string_of_data_type (input: data_type) : string =
  match input with
  | Num -> "Num"
  | Bool -> "Bool"
  | String -> "String"
  | Void -> "Void"
  | Object(id) -> "Object of type " ^ string_of_int id
  | List(t) -> "List of " ^ string_of_data_type t
  | Dict(k, v) -> "Dict of (" ^ string_of_data_type k ^ ": " ^ string_of_data_type v ^ ")"
  | Undetermined -> "Undetermined"
  | T(t) -> t

and string_of_aexpr (input: aexpr) : string =
  match input with
  | ANumLit(_, d) -> string_of_data_type d ^ " literal"
  | ABoolLit(_, d) -> string_of_data_type d ^ " literal"
  | AStringLit(_, d) -> string_of_data_type d ^ " literal"
  | AVal(_, d) -> "My AVal of type " ^ string_of_data_type d
  | AUnop(_, _, d) -> "My AUnop of type: " ^ string_of_data_type d
  | ABinop(_, _, _, d) -> "My ABinop of type: " ^ string_of_data_type d
  | AAssign(_, _, d) -> "My AAssign of type: " ^ string_of_data_type d
  | ACall(_, _, d) -> "My ACall of type: " ^ string_of_data_type d
  (* List Ops *)
  | AListDecl(_, d) -> "My AListDecl of type " ^ string_of_data_type d
  | AListInsert(_, _, _, d) -> "My AListInsert of type " ^ string_of_data_type d
  | AListPush(_, _, d) -> "My AListPush of type " ^ string_of_data_type d
  | AListRemove(_, _, d) -> "My AListRemove of type " ^ string_of_data_type d
  | AListPop(_, d) -> "My AListPop of type " ^ string_of_data_type d
  | AListDequeue(_, d) -> "My AListDequeue of type " ^ string_of_data_type d
  | AListLength(_, d) -> "My AListLength of type " ^ string_of_data_type d
  | AListGet(_, _, d) -> "My AListGet of type " ^ string_of_data_type d
  | AListSet(_, _, _, d) -> "My AListSet of type " ^ string_of_data_type d
  (* Dict Ops *)
  | ADictDecl(_, _, d) -> "My ADictDecl of type " ^ string_of_data_type d
  | ADictMem(_, _, d) -> "My ADictMem of type " ^ string_of_data_type d
  | ADictDelete (_, _, d) -> "My ADictDelete of type " ^ string_of_data_type d
  | ADictSize(_, d) -> "My ADictSize of type " ^ string_of_data_type d
  | ADictFind(_, _, d) -> "My ADictFind of type " ^ string_of_data_type d
  | ADictMap(_, _, _, d) -> "My ADictMap of type " ^ string_of_data_type d
  (* Obj Ops *)
  | AObjectAssign(_, _, _, d) -> "My AObjectAssign of type " ^ string_of_data_type d
  | AObjectField(_, _, d) -> "My AObjectField of type " ^ string_of_data_type d
  | _ -> raise (failwith("Could not get string of aexpr"))

and type_of_aexpr (input: aexpr) : data_type =
  match input with
  | ANumLit(_, d) -> d
  | ABoolLit(_, d) -> d
  | AStringLit(_, d) -> d
  | AVal(_, d) -> d
  | AUnop(_, _, d) -> d
  | ABinop(_, _, _, d) -> d
  | AAssign(_, _, d) -> d
  | ACall(_, _, d) -> d
  (* List Ops *)
  | AListDecl(_, d) -> d
  | AListInsert(_, _, _, d) -> d
  | AListPush(_, _, d) -> d
  | AListRemove(_, _, d) -> d
  | AListPop(_, d) -> d
  | AListDequeue(_, d) -> d
  | AListLength(_, d) -> d
  | AListGet(_, _, d) -> d
  | AListSet(_, _, _, d) -> d
  (* Dict Ops *)
  | ADictDecl(_, _, d) -> d
  | ADictMem(_, _, d) -> d
  | ADictDelete (_, _, d) -> d
  | ADictSize(_, d) -> d
  | ADictFind(_, _, d) -> d
  | ADictMap(_, _, _, d) -> d
  (* Obj Ops *)
  | AObjectAssign(_, _, _, d) -> d
  | AObjectField(_, _, d) -> d
  | _ -> raise (failwith("Could not determine type of aexpr"))

and is_object_list_decl (ae: aexpr) : bool =
  match ae with
    | AListDecl(_, typ) ->
      (match typ with
        | List(typ) ->
          (match typ with
            | Object(_) -> true
            | Undetermined -> true
            | _ -> false
          )
        | _ -> false
      )
    | _ -> false

(*
assign_add_map: After an assignment, add that variable to the local function (string->data type) environment
Params:
  env: Function environment with keys of variables and values of types corresponding to those variables
  aexpr (AAssign): Annotated assignment of id, annotated expression, data type of expression (and therefore variable)
Returns:
  env: The environment after the variable has been added to the map with its data type
*)
and assign_add_map (env: environment) (input: aexpr) =
  match input with
    | AAssign(id, _, d) ->
      if (NameMap.mem id env && (NameMap.find id env <> d)) then raise (failwith ("mismatched types for id " ^ id))
      else NameMap.add id d env
    | AObjectAssign(_, f, _, d) ->
      if (Hashtbl.mem field_hash f && (Hashtbl.find field_hash f <> d)) then raise (failwith ("mismatched types for field " ^ f))
      else let _ = Hashtbl.add field_hash f d in env
    | AVal(id, d) ->
      if (NameMap.mem id env && (NameMap.find id env <> d)) then raise (failwith ("mismatched types for object " ^ id))
    else NameMap.add id d env
    | _ -> raise (failwith ("Expected assign"))

(*
annotate_expr: 1st step in inference - annotate a variable, including the base cases of basic data types
Params:
  e: AST expression to be annotated
  env: Function environment with keys of variables and values of types corresponding to those variables
  aexpr (AAssign): Annotated assignment of id, annotated expression, data type of expression (and therefore variable)
Returns:
  aexpr: Annotated expression of the expression
*)
(* Step 1: Annotate Expressions *)
and annotate_expr (e: expr) (env: environment) : aexpr =
  match e with
  (* Base Cases *)
  | Num_lit(n) -> ANumLit(n, Num)
  | Bool_lit(b) -> ABoolLit(b, Bool)
  | String_lit(s) -> AStringLit(s, String)
  (* Variable -> Find variable in map, if it has not been declared yet throw an error *)
  | Id(x) -> if NameMap.mem x env
    then AVal(x, NameMap.find x env)
    else raise (failwith "Variable not defined")

  (* OBJECTS *)
  | ObjectField(o, f) -> if Hashtbl.mem field_hash f
    then let et = annotate_expr o env in AObjectField(et, f, Hashtbl.find field_hash f)
    else raise (failwith "Object field not defined")
  | ObjectAssign(o, f, e) ->
    let et1 = match o with
      | Id(x) -> if NameMap.mem x env
        then AVal(x, NameMap.find x env)
        else AVal(x, Object(-1))
      | _ -> annotate_expr o env
    and et2 = annotate_expr e env
    and new_type = gen_new_placeholder () in
    AObjectAssign(et1, f, et2, new_type)

  (* Binop -> Annotate the left and right sides of the binop first before combining into an annotated Binop *)
  | Binop(e1, op, e2) ->
    let et1 = annotate_expr e1 env
    and et2 = annotate_expr e2 env
    and new_type = gen_new_placeholder () in
    (match op with
      | Combine ->
        let ae1 = fst (infer env e1) and ae2 = fst (infer env e2) in
        let t1 = type_of ae1 and t2 = type_of ae2 in
        let x = match t1 with
          | List(u) -> u
          | _ -> raise (failwith "Not a list")
        and y = match t2 with
          | List(v) -> v
          | _ -> raise (failwith "Not a list")
        in let x_base = (base_type_of x 0) and y_base = (base_type_of y 0) in
          if snd x_base <> snd y_base then raise (failwith "mixed embedded lists")
          else
            let x_type = fst x_base and y_type = fst y_base in
              if ((x_type = Undetermined) && (y_type = Undetermined)) then
                ABinop(ae1, op, ae2, List(x))
              else if x_type = Undetermined then ABinop(ae1, op, ae2, List(y))
              else ABinop(ae1, op, ae2, List(x))
      | _ -> ABinop(et1, op, et2, new_type))
  (* Unop -> Annotate the right side before combining into an annotated Unop *)
  | Unop(op, e) ->
    let et = annotate_expr e env
    and new_type = gen_new_placeholder () in
    AUnop(op, et, new_type)
  (* Assign -> Annotate the right side before combining into an annotated Assign *)
  | Assign(id, e) ->
    let et = annotate_expr e env
    and new_type = gen_new_placeholder () in
    AAssign(id, et, new_type)

  (* LIST OPS *)
  | ListDecl(el) ->
    (match el with
      | [] -> AListDecl([], List(Undetermined))
      | _  -> let ael = List.map (fun x -> fst (infer env x)) el in
        let t = assert_list_type ael in
        AListDecl(ael, List(t))
    )
  | ListInsert(e1, e2, e3) ->
    let ae1 = fst (infer env e1) and ae2 = fst (infer env e2) and ae3 = fst (infer env e3) in
    AListInsert(ae1, ae2, ae3, gen_new_placeholder())
  | ListPush(e1, e2) ->
    let ae1 = fst (infer env e1) and ae2 = fst (infer env e2) in
    AListPush(ae1, ae2, gen_new_placeholder())
  | ListRemove(e1, e2) ->
    let ae1 = fst (infer env e1) and ae2 = fst (infer env e2) in
    AListRemove(ae1, ae2, gen_new_placeholder())
  | ListPop(e) ->
    let ae = fst (infer env e) in
    AListPop(ae, gen_new_placeholder())
  | ListDequeue(e) ->
    let ae = fst (infer env e) in
    AListDequeue(ae, gen_new_placeholder())
  | ListLength(e) ->
    let ae = fst (infer env e) in
    AListLength(ae, gen_new_placeholder())
  | ListGet(e1, e2) ->
    let ae1 = fst (infer env e1) and ae2 = fst (infer env e2) in
    AListGet(ae1, ae2, gen_new_placeholder())
  | ListSet(e1, e2, e3) ->
    let ae1 = fst (infer env e1) and ae2 = fst (infer env e2) and ae3 = fst (infer env e3) in
    AListSet(ae1, ae2, ae3, gen_new_placeholder())

  (* DICT OPS *)
  | DictDecl(tl) ->
    (match tl with
      | [] -> ADictDecl([], [], Dict(Undetermined, Undetermined))
      | _ ->
        let kl = List.map (fun x -> fst x) tl
        and vl = List.map (fun x -> snd x) tl
        in let akl = List.map (fun x -> fst (infer env x)) kl
        and avl = List.map (fun x -> fst (infer env x)) vl in
        let kt = assert_basic_list_type akl
        and vt = assert_list_type avl
        in ADictDecl(akl, avl, Dict(kt, vt))
    )
  | DictMem(e1, e2) ->
    let ae1 = fst (infer env e1) and ae2 = fst (infer env e2) in
    ADictMem(ae1, ae2, Bool)
  | DictDelete(e1, e2) ->
    let ae1 = fst (infer env e1) and ae2 = fst (infer env e2) in
    ADictDelete(ae1, ae2, Bool)
  | DictSize(e) ->
    let ae = fst (infer env e) in ADictSize(ae, Num)
  | DictFind(e1, e2) ->
    let ae1 = fst (infer env e1) and ae2 = fst (infer env e2) in
    ADictFind(ae1, ae2, gen_new_placeholder())
  | DictMap(e1, e2, e3) ->
    let ae1 = fst (infer env e1) and ae2 = fst (infer env e2) and ae3 = fst (infer env e3) in
    ADictMap(ae1, ae2, ae3, Bool)

  (* Call -> search to see if the function has already been annotated, if not call inference on it *)
  | Call(id, args) ->
    let _ =  if not (Hashtbl.mem func_order_hash id) then
      let new_id = gen_new_index() in
        (Hashtbl.add func_order_hash id new_id)
    in
    (* annotate the arguments *)
    let new_args = List.map (fun x -> fst(infer env x)) args in
    (* Function already inferred *)
    if Hashtbl.mem func_type_hash id then
      ACall(id, new_args, Hashtbl.find func_type_hash id)
    (* Function not inferred yet *)
    else if not (Hashtbl.mem fun_vis_rec id) then
      let _ = Hashtbl.add fun_vis_rec id 0 in
      (* Infer the argument types to pass into the environment of the function abbout to be inferred *)
      let params_list = List.map (fun x -> type_of (fst (infer env x))) args in
      (* Call inference on the function *)
      let _ = infer_types_function(Hashtbl.find func_decl_hash id) params_list in
      ACall(id, new_args, Hashtbl.find func_type_hash id)
    else
      ACall(id, new_args, Undetermined)
  | _ -> raise (failwith "Could not annotate expr")


(*
type_of: Get the type of an annotated expression
Params:
  ae: annotated expression to get the type of
Returns:
  t: the type of the annotated expression
*)
and type_of (ae: aexpr): data_type =
  match ae with
  | ANumLit(_, t) | ABoolLit(_, t) | AStringLit(_, t) -> t
  | AListDecl(_, t) -> t
  | ADictDecl(_, _, t) -> t
  | AVal(_, t) -> t
  | ABinop(_, _, _, t) -> t
  | AUnop(_, _, t) -> t
  | AAssign(_, _, t) -> t
  | ACall(_, _, t) -> t
  (* LIST OPS *)
  | AListInsert(_, _, _, t) -> t
  | AListPush(_, _, t) -> t
  | AListRemove(_, _, t) -> t
  | AListPop(_, t) -> t
  | AListDequeue(_, t) -> t
  | AListLength(_, t) -> t
  | AListGet(_, _, t) -> t
  | AListSet(_, _, _, t) -> t
  (* DICT OPS *)
  | ADictMem(_, _, t) -> t
  | ADictDelete(_, _, t) -> t
  | ADictSize(_, t) -> t
  | ADictFind(_, _, t) -> t
  | ADictMap(_, _, _, t) -> t
  (* OBJECT OPS *)
  | AObjectField(_, _, t) -> t
  | AObjectAssign(_, _, _, t) -> t
  | _ -> Undetermined

and base_type_of (d: data_type) (acc: int): (data_type * int) =
  match d with
  | Num -> Num, acc
  | Bool -> Bool, acc
  | String -> String, acc
  | Void -> Void, acc
  | Object(id) -> Object(id), acc
  | List(t) -> base_type_of t (acc + 1)
  | Dict(t1, t2) -> Dict(t1, t2), acc
  | Undetermined -> Undetermined, acc
  | T(t) -> T(t), acc

and get_list_type (d: data_type): data_type =
  match d with
  | List(t) -> t
  | _ -> raise (failwith "Expected list")

and get_basic_type (ael: aexpr list) =
  match ael with
    | [] -> List(Undetermined)
    | [l] -> type_of l
    | hd :: tail ->
      (match type_of hd with
        | Num -> Num
        | Bool -> Bool
        | String -> String
        | Void -> Void
        | Object(id) -> Object(id)
        | Dict(k, v) ->
          if k = Undetermined then get_basic_type tail else Dict(k, v)
        | List(t) ->
          if (fst (base_type_of t 0)) = Undetermined then get_basic_type tail else List(t)
        | Undetermined -> get_basic_type tail
        | T(t) -> T(t)
      )

and compare_list_type (x: aexpr) (t: data_type): int =
  let x_base = base_type_of (type_of x) 0
  and t_base = base_type_of t 0 in
  if snd x_base <> snd t_base then raise (failwith "mixed embedded lists: 1st check") else
    if (fst x_base = fst t_base || fst x_base = Undetermined) then 1 else 0

and assert_list_type (ael: aexpr list) =
  let t = get_basic_type ael in
  let same_type = List.map (fun x-> compare_list_type x t) ael in
  let not_same = List.mem 0 same_type in
  if not_same then raise (failwith "Mismatched list or dict key types") else t

and assert_basic_list_type (ael: aexpr list) =
  let t = type_of (List.hd ael) in
  let same_type = List.map (fun x-> (type_of x) = t) ael in
  let not_same = List.mem false same_type in
  if not_same then raise (failwith "Mismatched key types") else t

(*
collect_expr: collect constraints (ie 'a = 'b, 'b = Num) within an annotated expression
Also conveniently handles semantic analysis for binary operations down the line when types are mismatched
Params:
  ae: annotated expression to form constraints for
Returns:
  constraint_list: List of constraints, where each element is a data type equality of two types
*)
(* Step 2: Collect Constraints *)
and collect_expr (ae: aexpr) : (data_type * data_type) list =
  match ae with
  (* Literals are already fine *)
  | ANumLit(_) | ABoolLit(_) | AStringLit(_) | AListDecl(_) | ADictDecl(_) | AVal(_) | ACall(_) -> []
  | AObjectField(_) -> []
  | ABinop(ae1, op, ae2, t) ->
    let et1 = type_of ae1 and et2 = type_of ae2 in
    let opc = match op with
      (* Left and right side must be num, final return is a num *)
      | Add | Minus | Times | Divide -> [(et1, Num); (et2, Num); (t, Num)]
      (* Left and right side must be same type, final return is a boolean *)
      | Greater | Less | Geq | Leq | Eq | Neq -> [(et1, et2); (t, Bool)]
      (* Left and right side must be booleans, final return is a boolean *)
      | And | Or -> [(et1, Bool); (et2, Bool); (t, Bool)]
      (* Left and right side must be strings, final return is a string *)
      | Concat -> [(et1, String); (et2, String); (t, String)]
      (* Lists must be of same type *)
      | Combine -> []
    in
    (* Combine constraints of left side, right side, and binop constraints themselves *)
    (collect_expr ae1) @ (collect_expr ae2) @ opc
  | AUnop(op, ae, t) ->
    let et = type_of ae in
    let opc = match op with
      (* Right side must be a num, final return is a num *)
      | Neg -> [(et, Num); (t, Num)]
      | Not -> [(et, Bool); (t, Bool)]
    (* Combine constraints of right side and unop constraints themselves *)
    in (collect_expr ae) @ opc

  (* LIST OPS *)
  | AListInsert(ae1, ae2, ae3, t) ->
    let t1 = type_of ae1 and t2 = type_of ae2 and t3 = type_of ae3 in
    let x = match t1 with
      | List(u) -> u
      | _ -> raise (failwith "Not a list")
    in if x = Undetermined
      then [(t2, Num); (t, t3)]
      else [(x, t3); (t2, Num); (t, t3)]
  | AListPush(ae1, ae2, t) ->
    let t1 = type_of ae1 and t2 = type_of ae2 in
    let x = match t1 with
      | List(u) -> u
      | _ -> raise (failwith "Not a list")
    in if x = Undetermined
      then [(t, t2)] else [(x, t2); (t, t2)]
  | AListRemove(ae1, ae2, t) ->
    let t1 = type_of ae1 and t2 = type_of ae2 in
    let x = match t1 with
      | List(u) -> u
      | _ -> raise (failwith "Not a list")
    in [(t2, Num); (t, x)]
  | AListPop(ae, t) ->
    let t1 = type_of ae in
    let x = match t1 with
      | List(u) -> u
      | _ -> raise (failwith "Not a list")
    in [(t, x)]
  | AListDequeue(ae, t) ->
    let t1 = type_of ae in
    let x = match t1 with
      | List(u) -> u
      | _ -> raise (failwith "Not a list")
    in [(t, x)]
  | AListLength(ae, t) ->
    let t1 = type_of ae in
    let _ = match t1 with
      | List(u) -> u
      | _ -> raise (failwith "Not a list")
    in [(t, Num)]
  | AListGet(ae1, ae2, t) ->
    let t1 = type_of ae1 and t2 = type_of ae2 in
    let x = match t1 with
      | List(u) -> u
      | _ -> raise (failwith "Not a list")
    in [(t2, Num); (t, x)]
  | AListSet(ae1, ae2, ae3, t) ->
    let t1 = type_of ae1 and t2 = type_of ae2 and t3 = type_of ae3 in
    let x = match t1 with
      | List(u) -> u
      | _ -> raise (failwith "Not a list")
     in [(x, t3); (t2, Num); (t, t3)]
  (* DICT OPS *)
  | ADictMem(ae1, ae2, t) ->
    let t1 = type_of ae1 and t2 = type_of ae2 in
    let x = match t1 with
      | Dict(t, _) -> t
      | _ -> raise (failwith "Not a map")
    in [(x, t2); (t, Bool)]
  | ADictDelete(ae1, ae2, t) ->
    let t1 = type_of ae1 and t2 = type_of ae2 in
    let x = match t1 with
      | Dict(k, _) -> k
      | _ -> raise (failwith "Not a map")
    in [(x, t2); (t, Bool)]
  | ADictSize(ae, t) ->
    let aet = type_of ae in
    let _ = match aet with
      | Dict(_, _) -> ()
      | _ -> raise (failwith "Not a map")
    in [(t, Num)]
  | ADictFind(ae1, ae2, t) ->
    let t1 = type_of ae1 and t2 = type_of ae2 in
    let x, y = match t1 with
      | Dict(k, v) -> k, v
      | _ -> raise (failwith "Not a map")
    in [(x, t2); (t, y)]
  | ADictMap(ae1, ae2, ae3, t) ->
    let t1 = type_of ae1 and t2 = type_of ae2 and t3 = type_of ae3 in
    let x, y = match t1 with
      | Dict(k, v) -> k, v
      | _ -> raise (failwith "Not a map")
    in if x = Undetermined && y = Undetermined then
      [(t, Bool)] else [(x, t2); (y, t3); (t, Bool)]

  | AAssign(_, ae, lhs_type) ->
    (* lhs = variable name type (if reusing variable, it must be reused as same type, ie static inference) *)
    (* collect constraints of the right hand side, then type of left and right sides must be equal *)
    let rhs_type = type_of ae in
      (collect_expr ae) @ [(rhs_type, lhs_type)]
  | AObjectAssign(_, _, ae, lhs_type) ->
    let rhs_type = type_of ae in
      (collect_expr ae) @ [(rhs_type, lhs_type)]
  | _ -> raise (failwith "Not an aexpr in collect")

(* Step 3: Unification *)
and substitute (u: data_type) (x: id) (t: data_type) : data_type =
  match t with
  | Num | Bool | String | Void| Undetermined | Object(_) | List(_) | Dict(_) -> t
  | T(c) -> if c = x then u else t

and apply (subs: substitutions) (t: data_type) : data_type =
  List.fold_right (fun (x, u) t -> substitute u x t) subs t

and unify (constraints: (data_type * data_type) list) : substitutions =
  match constraints with
  | [] -> []
  | (x, y) :: xs ->
    (* generate substitutions of the rest of the list *)
    let t2 = unify xs in
    (* resolve the LHS and RHS of the constraints from the previous substitutions *)
    let t1 = unify_one (apply t2 x) (apply t2 y) in
    t1 @ t2

and unify_one (t1: data_type) (t2: data_type) : substitutions =
  match t1, t2 with
  | Num, Num | Bool, Bool | String, String | Undetermined, _ | _, Undetermined -> []
  | T(x), z | z, T(x) -> [(x, z)]
  | _ -> if t1 = t2 then []
    else raise (failwith "mismatched types")

and apply_expr (subs: substitutions) (ae: aexpr) (env: environment): (aexpr * environment) =
  match ae with
  | ABoolLit(b, t) -> ABoolLit(b, apply subs t), env
  | ANumLit(n, t) -> ANumLit(n, apply subs t), env
  | AStringLit(s, t) -> AStringLit(s, apply subs t), env
  | AVal(s, t) -> AVal(s, apply subs t), env
  | ABinop(e1, op, e2, t) ->
      ABinop(fst (apply_expr subs e1 env), op, fst (apply_expr subs e2 env), apply subs t), env
  | AUnop(op, e, t) -> AUnop(op, fst (apply_expr subs e env), apply subs t), env
  | AAssign(id, e, t) ->
    let apply_expr_ret = apply_expr subs e env in
    let ret_aexpr = AAssign(id, fst apply_expr_ret, apply subs t) in
    let _ = if type_of ret_aexpr = Void then
      raise(failwith "Cannot assign to a void function call") in
    let ret_env = assign_add_map (snd apply_expr_ret) ret_aexpr in
    ret_aexpr, ret_env
  (* OBJECT OPS *)
  | AObjectAssign(o, f, e, t) ->
    let apply_expr_ret = apply_expr subs e env in
    let ret_env1 = match o with
      | AVal(_, Object(_)) -> assign_add_map (snd apply_expr_ret) o
      | _ -> snd apply_expr_ret
    in let ret_aexpr = AObjectAssign(o, f, fst apply_expr_ret, apply subs t) in
    let ret_env2 = assign_add_map ret_env1 ret_aexpr in
    ret_aexpr, ret_env2
  | AObjectField(o, f, t) -> AObjectField(o, f, apply subs t), env
  | ACall(id, args, t) -> ACall(id, args, apply subs t), env
  | AListDecl(el, t) -> AListDecl(el, apply subs t), env
  | ADictDecl(kl, vl, t) -> ADictDecl(kl, vl, apply subs t), env
  (* LIST OPS *)
  | AListInsert(e1, e2, e3, t) ->
    let app = apply subs t in
    (* If the list is a variable, update the variable type after the insert *)
    let up_env = match e1 with
      | AVal(s, vt) ->
        let res = base_type_of vt 0 in
        if fst res = Undetermined then (set_iter_type s (List(app)) env) else env
      | AListGet(_, _, vt) ->
        let res = base_type_of vt 0 in
        if fst res = Undetermined then (handle_emb_list e1 (List(app)) env) else env
      | ADictFind(_, _, vt) ->
        let kt, vt = match vt with
          | Dict(k, v) -> k, v
          | _ -> raise (failwith "Expected map")
        in let res = base_type_of vt 0 in
        if fst res = Undetermined then (handle_emb_list e1 (Dict(kt, app)) env) else env
      | AObjectField(_, _, t) ->
        let res = base_type_of t 0 in
        if fst res = Undetermined then (handle_emb_list e1 (List(app)) env) else env
      | _ -> env
    in AListInsert(e1, e2, e3, app), up_env
  | AListPush(e1, e2, t) ->
    let app = apply subs t in
    (* If the list is a variable, update the variable type after the push *)
    let up_env = match e1 with
      | AVal(s, vt) ->
        let res = base_type_of vt 0 in
        if fst res = Undetermined then (set_iter_type s (List(app)) env) else env
      | AListGet(_, _, vt) ->
        let res = base_type_of vt 0 in
        if fst res = Undetermined then (handle_emb_list e1 (List(app)) env) else env
      | ADictFind(e, _, _) ->
        let vt = match type_of e with
          | Dict(_, v) -> v
          | _ -> raise (failwith "Expected map")
        in let res = base_type_of vt 0 in
        if fst res = Undetermined then (handle_emb_list e1 (List(app)) env) else env
      | AObjectField(_, _, t) ->
        let res = base_type_of t 0 in
        if fst res = Undetermined then (handle_emb_list e1 (List(app)) env) else env
      | _ -> env
    in AListPush(e1, e2, apply subs t), up_env
  | AListRemove(e1, e2, t) -> AListRemove(e1, e2, apply subs t), env
  | AListGet(e1, e2, t) -> AListGet(e1, e2, apply subs t), env
  | AListPop(e, t) -> AListPop(e, apply subs t), env
  | AListDequeue(e, t) -> AListDequeue(e, apply subs t), env
  | AListLength(e, t) -> AListLength(e, apply subs t), env
  | AListSet(e1, e2, e3, t) -> AListSet(e1, e2, e3, apply subs t), env
  (* DICT OPS *)
  | ADictMem(e1, e2, t) -> ADictMem(e1, e2, apply subs t), env
  | ADictDelete(e1, e2, t) -> ADictDelete(e1, e2, apply subs t), env
  | ADictSize(e, t) -> ADictSize(e, apply subs t), env
  | ADictFind(e1, e2, t) -> ADictFind(e1, e2, apply subs t), env
  | ADictMap(e1, e2, e3, t) ->
    let app = apply subs t
    and t2 = type_of e2 and t3 = type_of e3 in
    let up_env = match e1 with
      | AVal(s, vt) ->
        let k = match vt with
          | Dict(x, _) -> x
          | _ -> raise (failwith "Expected a dict")
        in if k = Undetermined then set_iter_type s (Dict(t2, t3)) env else env
      | AListGet(e1, _, t) ->
        let k = match t with
          | Dict(x, _) -> x
          | _ -> raise (failwith "Expected a dict")
        in if k = Undetermined then (handle_emb_list e1 (List(Dict(t2, t3))) env) else env
      | AObjectField(_, s, t) ->
        let k = match t with
          | Dict(x, _)-> x
          | _ -> raise(failwith "Expected a dict")
        in if k = Undetermined then set_iter_type s (Dict(t2, t3)) env else env
      | _ -> env
    in ADictMap(e1, e2, e3, app), up_env
  | _ -> raise (failwith("Invalid aexpr in apply_aexpr"))

(*
handle_emb_list: Changes the type of an undetermined embedded list (ie [[], []]) after a push or insert
Params:
  aexpr: The current layer of expression. The total form should be a number of AListGet's with an AVal at the center
  t: The data_type thus far of the list, embeds more lists recursively
  env: the local function environment where the type of the list will be altered
Returns:
  env: The updated environment
*)
and handle_emb_list (ae: aexpr) (t: data_type) (env: environment): environment =
  match ae with
    | AVal(s, _) ->
      let temp = NameMap.remove s env in
      NameMap.add s t temp
    | AListGet(e1, _, _) ->
      handle_emb_list e1 (List(t)) env
    | ADictFind(e1, _, _) ->
      let res = type_of e1 in
      let kt = match res with
        | Dict(k, _) -> k
        | _ -> raise (failwith "Expected dict")
      in handle_emb_list e1 (Dict(kt, t)) env
    | AObjectField(_, s, _) ->
      let _ = Hashtbl.remove field_hash s in
      let _ = Hashtbl.add field_hash s t
      and temp = NameMap.remove s env in
      NameMap.add s t temp
    | _ -> raise (failwith "Expected list or map")

(*
set_iter_type: general helper function to remove then re-add a type for a variable
Params:
  s: The variable name
  t: The data_type to be inserted
  env: the local function environment where the type of the list will be altered
Returns:
  env: The updated environment after s has been set to type t
*)
and set_iter_type (s: string) (t: data_type) (env: environment): environment =
  let _ = match Hashtbl.mem field_hash s with
    | true ->
      let _ = Hashtbl.remove field_hash s in
      let _ = Hashtbl.add field_hash s t in ()
    | _ -> ()
  in let temp = NameMap.remove s env in
  NameMap.add s t temp

(*
infer: Infer the annotated expression form of an expression by calling HM type inference
Params:
  env: the local function environment, used in case e is an assignment or contains an Id
  e: The expression to perform type inference on
Returns:
  aexpr: annotated version of the expression
  env: local function environment
*)
and infer (env: environment) (e: expr) : (aexpr * environment) =
  let annotated_expr = annotate_expr e env in
  let constraints = collect_expr annotated_expr in
  let subs = unify constraints in
  (* reset the type counter after completing inference *)
  type_variable := (Char.code 'a');
  let ret = apply_expr subs annotated_expr env in
  if logging then
   let _ = print_endline (string_of_aexpr (fst ret)) in ret else ret

(*
infer: Infer the annotated expression form of an statement
Params:
  env: the local function environment, used in case e is an assignment or contains an Id
  stmt: The statement to be annotated
Returns:
  astmt: annotated version of the statement
  env: local function environment, updated to include possible changes due to assignments
*)
and infer_stmt (env: environment) (stmt: stmt) (func_name: string): (astmt * environment) =
  match stmt with
  (* Expr -> infer the expression *)
  | Expr(e) -> let ret = infer env e in AExpr(fst ret), snd ret
  (* Return -> infer the expression being returned *)
  | Print(e) -> let ret = infer env e in APrint(fst ret), snd ret
  | Return(e) ->
    (match e with
      | Ast.Noexpr -> AReturn(Sast.ANoexpr(Void)), env
      | _ ->
        let ret = infer env e in
        let d = type_of (fst ret) in
        if (Hashtbl.mem func_type_hash func_name) then
          (* Functions must return same type *)
          let _ = log ("[infer_stmt] looking up " ^ func_name ^ " in func_type_hash") in
          if (Hashtbl.find func_type_hash func_name <> d) then
            raise (failwith ("[infer_stmt] mismatched types of function " ^ func_name))
          else
            AReturn(fst ret), snd ret
        (* If function has not been inferred before, add the return type of the function *)
        else
          let _ = Hashtbl.add func_type_hash func_name d in
            AReturn(fst ret), snd ret
    )
  | Break -> ABreak, env
  | Continue -> AContinue, env
  | ObjectInit(obj_list) ->
    let new_env = List.fold_left (fun acc obj_name -> (NameMap.add obj_name (Object(-1)) acc)) env obj_list in
    AObjectInit(List.map (fun x -> (x, Object(-1))) obj_list), new_env
  | While(e, s) ->
    (* Infer the condition *)
    let inf = fst (infer env e) in
    let _ = match type_of inf with
      | Bool -> "ok"
      | _ -> raise (failwith ("expected boolean in if statement"))
    (* unwrap the statement list from the block *)
    and bl = match s with
      | Block(sl) -> sl
      | _ -> raise (failwith ("expected block"))
    (* Call inference on the block *)
    in let abl = fst (map_infer_stmt bl env [] func_name) in
    (* let _ = print_endline (string_of_data_type (type_of inf)) in *)
    AWhile(inf, ABlock(abl)), env
  | If(e, s1, s2) ->
    (* infer the condition *)
    let inf = fst (infer env e) in 
    let _ = match type_of inf with
      | Bool -> "ok"
      | _ -> raise (failwith ("expected boolean in if statement"))
    (* unwrap the statement lists from the blocks *)
    and bl1 = match s1 with
      | Block(sl) -> sl
      | _ -> raise (failwith ("expected block"))
    and bl2 = match s2 with
      | Block(sl) -> sl
      | _ -> raise (failwith ("expected block"))
    (* infer the blocks *)
    in let abl1 = fst (map_infer_stmt bl1 env [] func_name)
    in let abl2 = fst (map_infer_stmt bl2 env [] func_name)
    in AIf(inf, ABlock(abl1), ABlock(abl2)), env
  | ForRange(e1, e2, e3, s) ->
    (* infer the from-to expression pair *)
    let inf1_ae, inf1_env = infer env e1 in
    let inf2_ae, inf2_env = infer inf1_env e2 in
    let inf3_ae, _ = infer inf2_env e3 in
    let _ = match inf1_ae with
      | AAssign(_) -> (match type_of inf1_ae with
        | Num -> "ok"
        | _ -> raise (failwith ("assignment in for loop must be a number")))
      | _ -> raise (failwith ("expected assignment in for statement"))
    and _ = (match type_of inf2_ae with
      | Num -> "ok"
      | _ -> raise (failwith ("end in for statement must be a number")))
    and _ = (match type_of inf3_ae with
      | Num -> "ok"
      | _ -> raise (failwith ("by in for statement must be a number")))
    (* unwrap the statement list from the block *)
    and bl = match s with
      | Block(sl) -> sl
      | _ -> raise (failwith ("expected block"))
    (* infer the blocks *)
    in let abl = fst (map_infer_stmt bl inf1_env [] func_name)
    in AForRange(inf1_ae, inf2_ae, inf3_ae, ABlock(abl)), env
  | ForIn(e1, e2, s) ->
    (* infer the for-in expression pair *)
    let vname = (match e1 with
      | Id(s) -> s
      | _ -> raise (failwith "First term in For In must be a variable"))
    and inf2 = fst (infer env e2)
    in let t = (match type_of inf2 with
      | Dict(t, _) | List(t) -> t
      | _ -> raise (failwith "Second term in For In must be an iterable"))
    in let new_env = NameMap.add vname t env
    (* unwrap the statement list from the block *)
    in let bl = match s with
      | Block(sl) -> sl
      | _ -> raise (failwith ("expected block"))
    (* infer the block *)
    in let abl = fst (map_infer_stmt bl new_env [] func_name)
    in AForIn(AVal(vname, t), inf2, ABlock(abl)), new_env
  | _ -> raise (failwith ("unexpected statement"))

(*
map_infer_stmt: for a given statment, infer the type of it while maintaining the scope's name map
params:
  stmt_list: list of statements from a function's body
  env: Name map with keys of IDs and values of the IDs data type
  astmt_list: final return list of annotated statements
  f_name: the name of the function the statement is in - used to handle returns and adding to the function type map
returns:
  astmt_list: annotated version of the stmt list passed in
  env: the name map with keys of IDs and values of the IDs data type
    -only used for the recursion
*)
and map_infer_stmt (stmt_list: stmt list) (env: environment) (astmt_list: astmt list) (f_name: string): (astmt list * environment) =
  (* base case - all statements have been inferred *)
  match stmt_list with
      [] -> astmt_list, env
    | head :: tail ->
      (* infer the first statement, concat the annotated version to the accumulator *)
      let inf = infer_stmt env head f_name in
      let new_astmt_list = astmt_list @ [fst inf]
      (* update the environment *)
      and new_env = snd inf in
      map_infer_stmt tail new_env new_astmt_list f_name

(*
add_tuple_to_env: adds an id tuple to a local function environment
params:
  env: the data type map to add to
  tuple: the (data type, string) tuple to be added
returns:
  env: the environment after it has been updated
*)
and add_tuple_to_env (env: environment) (tuple: (data_type * string)) =
  NameMap.add (snd tuple) (fst tuple) env

(*
get_param_types: gets the parameter types of a function
params:
  formals: list of Ids that are the parameters
  types: In the same order, the list of the types of those paraeters
  env: The initial starting environment (should be NameMap.empty but passed just in case)
returns:
  final_env: the environment after the tuples have been added
  tuple_list: the (data_type * string) formals list to be input in aformals
*)
and get_param_types (formals: string list) (types: data_type list) (env: environment) =
  let tuple_list = List.map2 (fun a b -> (a, b)) types formals in
  let final_env = List.fold_left add_tuple_to_env env tuple_list in
  final_env, tuple_list

(*
infer_types_function: Calls inference on a function, given its declaration
params:
  func: The function declaration, as taken from the AST's list of fdecls
  param_types: The types of the parameters, as inferred from the first time the function is called
returns:
  abody: The annotated list of statements that constitute the annotated function's body
*)
and infer_types_function (func: func_decl) (param_types: data_type list) =
  (* Load the function parameters into the function environment, get aformals in the process *)
  let env_and_aformals = get_param_types func.formals param_types NameMap.empty in
  let func_env = fst env_and_aformals
  and aforms = snd env_and_aformals in
  (* Call inference on the function body *)
  let ret = map_infer_stmt func.body func_env [] func.fname in
  (* Add the function body and function formals to the global body and formal tables respectively *)
  let _ = Hashtbl.add func_body_hash func.fname (fst ret)
  and _ = Hashtbl.add func_formals_hash func.fname aforms
  and _ = if not (Hashtbl.mem func_type_hash func.fname)
    then Hashtbl.add func_type_hash func.fname Void in
  let print_env k v =
    log (k ^ ": " ^ string_of_data_type v)
    in let _ = NameMap.iter print_env (snd ret) in
    let _ = Hashtbl.add func_env_hash func.fname (snd ret) in
  fst ret


and handle_undet_expr (e: Sast.aexpr) (name: string) =
  match e with
    | ABinop(ae1, op, ae2, typ) ->
      let hae1 = handle_undet_expr ae1 name and hae2 = handle_undet_expr ae2 name in
      let t1 = type_of hae1 and t2 = type_of hae2 in
      let _ = match op with
        | Add | Minus | Times | Divide ->
          if not (t1 = Num && t2 = Num) then
            raise (failwith "Must be a Num")
          else ()
        | Greater | Less | Geq | Leq | Eq | Neq ->
          if not (t1 = t2) then
            raise (failwith "Must be a same type")
          else ()
        | And | Or ->
          if not (t1 = Bool && t2 = Bool) then
            raise (failwith "Must be a Bool")
          else ()
        | Concat ->
          if not (t1 = String && t2 = String) then
            raise (failwith "Must be a same type")
          else ()
        | Combine -> ()
      in ABinop(hae1, op, hae2, typ)
    | AUnop(op, ae, typ) -> 
      let hae = handle_undet_expr ae name in
      let t = type_of hae in
      let _ = match op with
        | Neg ->
          if not (t = Num) then
            raise (failwith "Must be a Num")
          else ()
        | Not ->
          if not (t = Bool) then
            raise (failwith "Must be a Bool")
          else ()
      in AUnop(op, hae, typ)
    | AAssign(id, ae, typ) ->
      if fst (base_type_of typ 0) = Undetermined then
        let new_ae = handle_undet_expr ae name in
        let new_typ = type_of new_ae in
        let new_typ = if fst (base_type_of new_typ 0) = Undetermined then
          let _ = log ("[undet] " ^ string_of_aexpr e) in
          let _ = log ("[undet] " ^ string_of_data_type (NameMap.find id (Hashtbl.find func_env_hash name))) in
          NameMap.find id (Hashtbl.find func_env_hash name)
          else new_typ
        in AAssign(id, new_ae, new_typ)
      else
        AAssign(id, handle_undet_expr ae name, typ)
    | ACall(id, aexpr_list, typ) ->
      let new_aexpr_list = List.map (fun x -> handle_undet_expr x name) aexpr_list in
      let new_typ = match typ with
        | Undetermined -> Hashtbl.find func_type_hash id
        | _ -> typ
      in ACall(id, new_aexpr_list, new_typ)
    | AListDecl(aexpr_list, typ) ->
      let new_aexpr_list = List.map (fun x -> handle_undet_expr x name) aexpr_list in
      AListDecl(new_aexpr_list, typ)
    | AListInsert(ae1, ae2, ae3, typ) ->
      AListInsert(handle_undet_expr ae1 name, handle_undet_expr ae2 name, handle_undet_expr ae3 name, typ)
    | AListPush(ae1, ae2, typ) ->
      AListPush(handle_undet_expr ae1 name, handle_undet_expr ae2 name, typ)
    | AListRemove(ae1, ae2, typ) ->
      AListRemove(handle_undet_expr ae1 name, handle_undet_expr ae2 name, typ)
    | AListPop(ae, typ) ->
      AListPop(handle_undet_expr ae name, typ)
    | AListDequeue(ae, typ) ->
      AListDequeue(handle_undet_expr ae name, typ)
    | AListLength(ae, typ) ->
      AListLength(handle_undet_expr ae name, typ)
    | AListGet(ae1, ae2, typ) ->
      AListGet(handle_undet_expr ae1 name, handle_undet_expr ae2 name, typ)
    | AListSet(ae1, ae2, ae3, typ) ->
      AListSet(handle_undet_expr ae1 name, handle_undet_expr ae2 name, handle_undet_expr ae3 name, typ)
    | ADictDecl(ael1, ael2, typ) -> 
      let map_handle = (fun x -> handle_undet_expr x name) in
      ADictDecl(List.map map_handle ael1, List.map map_handle ael2, typ)
    | ADictMem(ae1, ae2, typ) ->
      ADictMem(handle_undet_expr ae1 name, handle_undet_expr ae2 name, typ)
    | ADictDelete(ae1, ae2, typ) ->
      ADictDelete(handle_undet_expr ae1 name, handle_undet_expr ae2 name, typ)
    | ADictFind(ae1, ae2, typ) ->
      ADictFind(handle_undet_expr ae1 name, handle_undet_expr ae2 name, typ)
    | ADictSize(ae, typ) ->
      ADictSize(handle_undet_expr ae name, typ)
    | ADictMap(ae1, ae2, ae3, typ) ->
      ADictMap(handle_undet_expr ae1 name, handle_undet_expr ae2 name, handle_undet_expr ae3 name, typ)
    | AObjectField(ae, s, typ) ->
      AObjectField(handle_undet_expr ae name, s, typ)
    | AObjectAssign(ae1, s, ae2, typ) ->
      AObjectAssign(handle_undet_expr ae1 name, s, handle_undet_expr ae2 name, typ)
    | _ -> e



and handle_undet_stmt (s: Sast.astmt) (name: string) =
  match s with
    | AExpr(aexpr) -> AExpr(handle_undet_expr aexpr name)
    | ABlock(astmt_list) -> ABlock(List.map (fun x-> handle_undet_stmt x name) astmt_list)
    | AReturn(aexpr) -> AReturn(handle_undet_expr aexpr name)
    | AWhile(aexpr, astmt) -> AWhile(handle_undet_expr aexpr name, handle_undet_stmt astmt name)
    | AForIn(aexpr1, aexpr2, astmt) ->
      AForIn(handle_undet_expr aexpr1 name, handle_undet_expr aexpr2 name, handle_undet_stmt astmt name)
    | AForRange(aexpr1, aexpr2, aexpr3, astmt) ->
      AForRange(handle_undet_expr aexpr1 name, handle_undet_expr aexpr2 name,
                handle_undet_expr aexpr3 name, handle_undet_stmt astmt name)
    | AIf(aexpr, astmt1, astmt2) ->
      AIf(handle_undet_expr aexpr name, handle_undet_stmt astmt1 name, handle_undet_stmt astmt2 name)
    | APrint(expr) -> APrint(handle_undet_expr expr name)
    | _ -> s


and handle_undet_func (afdecl: Sast.afunc_decl) =
  let new_afunc_decl = {
    afname = afdecl.afname;
    aformals = afdecl.aformals;
    abody = List.map (fun x -> handle_undet_stmt x afdecl.afname) afdecl.abody;
    return = afdecl.return;
  } in new_afunc_decl

and handle_undet (sast: Sast.aprogram) =
  List.map handle_undet_func sast







(* ==================== OBJECTS ==================== *)

(* Maps each object to its list of fields: <obj, [field1, field2, field3]> *)
and obj_to_fields_hash = Hashtbl.create 42

(* Maps each object to its type id: <obj, id> *)
and obj_to_id_hash = Hashtbl.create 42

(* Maps each type id to its list of fields <id, [field1, field2, field3]> *)
and id_to_fields_hash = Hashtbl.create 42

(* Maps each type id to its list of fields with types <id, [(field1, Num), (field2, Bool), (field3, String)]> *)
and id_to_fields_typified_hash = Hashtbl.create 42

(* Maps function names to the list of object names that are equivalent to the function return *)
and fname_to_obj_equality_hash = Hashtbl.create 42

(* Maps function names to whether or not they have been visisted by obj_collect *)
and fname_to_visited_hash = Hashtbl.create 42

and obj_inserted = ref false

and obj_seen = ref false

and curr_id = ref 0

and gen_new_obj_id () : int =
  let n1 = !curr_id in
  let _ = incr curr_id in n1

(*
[ENTRY POINT]

infer_objs: Entry point into object inference algorithm
params:
  sast: the program to run object inference on
returns:
  a tuple of (afdecls, obj_fields_hash):
    afdecls: the augmented sast with object ids
    id_to_fields_hash: the map of ids to lists of fields
*)
and infer_objs (sast: Sast.aprogram) =
  (* Tracks objects that are of the same type: <obj1, obj2> *)
  let obj_equality_sets = obj_collect (Hashtbl.find afunc_decl_hash "main") [] in

  let _ = log "[infer_objs] obj_equality_sets" in
  let _ = print_set_list obj_equality_sets in
  let _ = print_obj_to_fields_hash() in

  let _ = obj_unify obj_equality_sets in

  let _ = print_obj_to_id_hash() in
  let _ = print_id_to_fields_hash() in

  let afdecls = obj_apply sast in
  let _ = typify_fields_hash() in
  let _ = print_id_to_fields_typified_hash() in
  (afdecls, id_to_fields_typified_hash)





(* ==================== OBJECT UTILS ==================== *)

and remove_dups lst = match lst with
  | [] -> []
  | h::t -> h::(remove_dups (List.filter (fun x -> x<>h) t))

and typify_fields_hash () =
  Hashtbl.iter (fun id fields ->
    let new_fields = List.map (fun field -> (field, Hashtbl.find field_hash field)) fields in
    let new_unique_fields = remove_dups new_fields in
    Hashtbl.add id_to_fields_typified_hash id new_unique_fields
  ) id_to_fields_hash

and add_field_to_obj (field: string) (obj: string) =
  let _ = log ("Adding " ^ field ^ " to object called " ^ obj) in
  let field_list =
    if Hashtbl.mem obj_to_fields_hash obj
    then Hashtbl.find obj_to_fields_hash obj
    else []
  in
  let new_field_list =
    if List.mem field field_list
    then field_list
  else field_list @ [field]
  in
  Hashtbl.add obj_to_fields_hash obj new_field_list

(*
add_new_equality: Inserts an equality of two objects into the equality set list
*)
and add_new_equality (eq: ObjectTypeSet.t list) (obj1: string) (obj2: string) : (ObjectTypeSet.t list) =
  let _ = log ("[add_new_equality] adding a new equality between " ^ obj1 ^ " and " ^ obj2) in
  let _ = log ("[add_new_equality] current set list looks like: ") in
  let _ = print_set_list eq in
  obj_inserted := false;
  let ret = List.map (fun (obj_set: ObjectTypeSet.t) ->
    let _ = log ("[add_new_equality] considering adding " ^ obj1 ^ "," ^ obj2 ^ " into the following set:") in
    let _ = print_set obj_set in
    let new_obj_set =
      if (ObjectTypeSet.mem obj1 obj_set && not (ObjectTypeSet.mem obj2 obj_set))
      then (obj_inserted := true; let _ = log ("[add_new_equality] adding " ^ obj2 ^ " into the above set") in ObjectTypeSet.add obj2 obj_set)
      else obj_set
    in
    let new_obj_set =
      if (ObjectTypeSet.mem obj2 obj_set && not (ObjectTypeSet.mem obj1 obj_set))
      then (obj_inserted := true; let _ = log ("[add_new_equality] adding " ^ obj1 ^ " into the above set") in ObjectTypeSet.add obj1 new_obj_set)
      else new_obj_set
    in
    new_obj_set
  ) eq in
  if !obj_inserted = false (*  if objected_inserted is equal to false *)
  then
    let _ = log ("[add_new_equality] adding a completely new set") in
    let new_set = ObjectTypeSet.add obj1 ObjectTypeSet.empty in
    let new_set = ObjectTypeSet.add obj2 new_set in
    let _ = print_set_list (ret @ [new_set]) in
    ret @ [new_set]
  else ret

and add_lone_equality (eq: ObjectTypeSet.t list) (obj: string) : (ObjectTypeSet.t list) =
  obj_seen := false;
  let _ = List.iter (fun obj_set ->
    if ObjectTypeSet.mem obj obj_set
    then (obj_seen := true; ())
    else ()
  ) eq
  in
  if !obj_seen = false
  then let new_set = ObjectTypeSet.add obj ObjectTypeSet.empty in
  eq @ [new_set]
  else eq

(*
get_arg_index: Returns the index of an argument in a call to a function
*)
and get_arg_index (args: aexpr list) (arg: string) : (int) =
  let rec get_idx key ael idx =
    match ael with
      | [] -> let _ = log "[get_arg_index] Base case" in raise (failwith "Arg not found in function")
      | head :: tail ->
        match head with
          | AVal(str, _) ->
            let _ = log ("[get_arg_index] Arg of name " ^ str ^ " found in the ACall, matching against " ^ key) in
            if str = key
            then idx
            else get_idx key tail (idx + 1)
          | _ -> raise (failwith("Argument not found in formals list"))
  in get_idx arg args 0

(*
get_obj_name_from_aexpr: Returns the object name from an aexpr or "" if it is not an object
*)
and get_obj_name_from_aexpr (ae: aexpr) : string =
  match ae with
    | AVal(name, typ) ->
      (match typ with
        | Object(_) -> name
        | _ -> ""
      )
    | AObjectField(_, field, typ) ->
      (match typ with
        | Object(_) -> field
        | _ -> ""
      )
    | _ -> ""

and get_list_type_from_aexpr (ae: aexpr) : int =
  let list_name = get_list_name_from_aexpr ae in
  let encoded_list_name = encode_list_name list_name in
  if Hashtbl.mem obj_to_id_hash encoded_list_name
  then Hashtbl.find obj_to_id_hash encoded_list_name
  else -1 (* Not a list of objects *)

and get_list_type_from_listdecl (ae: aexpr) : data_type =
  match ae with
    | AListDecl(ae_list, typ) ->
      if List.length ae_list = 0
      then typ
      else List(type_of_aexpr (List.hd ae_list))
    | _ -> raise (failwith("Attempting to get List type from not a ListDecl"))

and get_list_name_from_aexpr (annotated_expr: aexpr) : string =
  match annotated_expr with
    | AVal(name, typ) ->
      (match typ with
        | List(_) -> name
        | _ -> ""
      )
    | AListInsert(ae, _, _, _) -> get_list_name_from_aexpr ae
    | AListPush(ae, _, _) -> get_list_name_from_aexpr ae
    | AListRemove(ae, _, _) -> get_list_name_from_aexpr ae
    | AListPop(ae, _) -> get_list_name_from_aexpr ae
    | AListDequeue(ae, _) -> get_list_name_from_aexpr ae
    | AListLength(ae, _) -> get_list_name_from_aexpr ae
    | AListGet(ae, _, _) -> get_list_name_from_aexpr ae
    | AListSet(ae, _, _, _) -> get_list_name_from_aexpr ae
    | _ -> ""

and encode_list_name (list_name: string) : string =
  "." ^ list_name

and inject_list_decl_type (ae: aexpr) (new_typ: data_type) : aexpr =
  match ae with
    | AListDecl(contents, _) -> AListDecl(contents, new_typ)
    | _ -> raise (failwith("Injecting: Not an AListDecl"))

and encode_dict_name (dict_name: string) : string =
  "," ^ dict_name

and get_dict_name_from_aexpr (annotated_expr: aexpr) : string =
  match annotated_expr with
    | AVal(name, typ) ->
      (match typ with
        | Dict(_, _) -> name
        | _ -> ""
      )
    | ADictMem(ae, _, _) -> get_dict_name_from_aexpr ae
    | ADictDelete(ae, _, _) -> get_dict_name_from_aexpr ae
    | ADictSize(ae, _) -> get_dict_name_from_aexpr ae
    | ADictFind(ae, _, _) -> get_dict_name_from_aexpr ae
    | ADictMap(ae, _, _, _) -> get_dict_name_from_aexpr ae
    | _ -> ""

and get_dict_val_type_from_aexpr (ae: aexpr) : int =
  let dict_name = get_dict_name_from_aexpr ae in
  let _ = log ("aexpr is " ^ string_of_aexpr ae ^ " and dict_name is " ^ dict_name) in
  if Hashtbl.mem obj_to_id_hash dict_name
  then Hashtbl.find obj_to_id_hash dict_name
  else -1 (* Not a list of objects *)

(* Binds an object name to a function return type, adds and equality if applicable, and returns the set list *)
(* [Magic] *)
and insert_obj_into_fname_to_obj_equality_hash (acc: ObjectTypeSet.t list) (fname: string) (obj: string) : (ObjectTypeSet.t list) =
  let curr_list =
    if Hashtbl.mem fname_to_obj_equality_hash fname
    then Hashtbl.find fname_to_obj_equality_hash fname
    else []
  in
  let acc =
    if List.length curr_list >= 1
    then add_new_equality acc obj (List.hd curr_list)
    else acc
  in
  let new_list =
    if List.mem obj curr_list
    then curr_list
    else curr_list @ [obj]
  in
  let _ = Hashtbl.add fname_to_obj_equality_hash fname new_list in
  acc





(* ==================== OBJECT COLLECT ==================== *)

and obj_collect_fields_aexpr (ae: aexpr) =
  match ae with
    (* Base Cases *)
    | AObjectField(ae, field, _) ->
      add_field_to_obj field (get_obj_name_from_aexpr ae)
    | AObjectAssign(ae, field, iae, _) ->
      let _ = obj_collect_fields_aexpr iae in
      add_field_to_obj field (get_obj_name_from_aexpr ae)
    (* Recursive Cases *)
    | ABinop(ae1, _, ae2, _) ->
      let _ = obj_collect_fields_aexpr ae1 in
      let _ = obj_collect_fields_aexpr ae2 in
      ()
    | AUnop(_, ae, _) ->
      obj_collect_fields_aexpr ae
    | AAssign(_, ae, _) ->
      obj_collect_fields_aexpr ae
    | ACall(_, ae_list, _) ->
      List.iter (fun ae -> obj_collect_fields_aexpr ae) ae_list
    | _ -> ()

(*
obj_collect_fields_astmt: If the given astmt contains an AObjectField, record it in obj_to_fields_hash
params:
  annotated_stmt: an astmt to collect fields from
returns:
  unit
*)
and obj_collect_fields_astmt (annotated_stmt: astmt) =
  match annotated_stmt with
    | ABlock(astmt_list) -> List.iter obj_collect_fields_astmt astmt_list
    | AExpr(ae) -> obj_collect_fields_aexpr ae
    | AReturn(ae) -> obj_collect_fields_aexpr ae
    | AWhile(ae, astmt) ->
      let _ = obj_collect_fields_aexpr ae in
      let _ = obj_collect_fields_astmt astmt in
      ()
    | AForIn(ae1, ae2, astmt) ->
      let _ = obj_collect_fields_aexpr ae1 in
      let _ = obj_collect_fields_aexpr ae2 in
      let _ = obj_collect_fields_astmt astmt in
      ()
    | AForRange(ae1, ae2, ae3, astmt) ->
      let _ = obj_collect_fields_aexpr ae1 in
      let _ = obj_collect_fields_aexpr ae2 in
      let _ = obj_collect_fields_aexpr ae3 in
      let _ = obj_collect_fields_astmt astmt in
      ()
    | AIf(ae, astmt1, astmt2) ->
      let _ = obj_collect_fields_aexpr ae in
      let _ = obj_collect_fields_astmt astmt1 in
      let _ = obj_collect_fields_astmt astmt2 in
      ()
    | APrint(ae) -> obj_collect_fields_aexpr ae
    | _ -> ()

(*
obj_collect_equalities_param: Collects the param equality for one param in a function call
*)
and obj_collect_equalities_param (fname: string) (args: aexpr list) (arg: aexpr) (acc: ObjectTypeSet.t list) : (ObjectTypeSet.t list) =
  match arg with
    | AVal(arg_name, typ) ->
      (* arg_name is the name of the object that is passed into the function call *)
      (* we find the index of arg_name in the function call, and get the corresponding param name in the afdecl *)
      (* then we add a new equality for them *)
      (match typ with
        | Object(_) ->
          let afdecl = Hashtbl.find afunc_decl_hash fname in
          let index = get_arg_index args arg_name in
          let param = List.nth afdecl.aformals index in
          let param_name = snd param in
          let acc = add_new_equality acc param_name arg_name in
          acc
        | _ -> acc
      )
    | _ -> acc

(*
obj_collect_equalities_params: Collects all param equalities for one function call
*)
and obj_collect_equalities_params (fname: string) (args: aexpr list) (acc: ObjectTypeSet.t list) : (ObjectTypeSet.t list) =
  List.fold_left (fun acc arg -> obj_collect_equalities_param fname args arg acc) acc args

and get_obj_or_list_or_map_name_from_aexpr (ae: aexpr) : string =
  let obj_name = get_obj_name_from_aexpr ae in
  if obj_name = ""
  then (
    let list_name = get_list_name_from_aexpr ae in
    if list_name = ""
    then (
      let dict_name = get_dict_name_from_aexpr ae in
      if dict_name = ""
      then ""
    else encode_dict_name dict_name
    )
    else encode_list_name list_name
  )
  else obj_name

(*
obj_collect_equalities_binop: Collects the equality from `obj1 == obj2`
*)
and obj_collect_equalities_binop (acc: ObjectTypeSet.t list) (lhs: aexpr) (op: binop) (rhs: aexpr) : (ObjectTypeSet.t list) =
  match op with
    | Eq ->
      (match type_of rhs with
        | Object(_) ->
          let lname = get_obj_or_list_or_map_name_from_aexpr lhs in
          let rname = get_obj_or_list_or_map_name_from_aexpr rhs in
          if (lname <> "") && (rname <> "")
          then add_new_equality acc lname rname
          else acc
        | _ -> acc
      )
    | _ -> acc

and obj_collect_equalities_list (acc: ObjectTypeSet.t list) (lst: aexpr) (obj: aexpr) =
  let list_name = get_list_name_from_aexpr lst in
  if list_name = ""
  then acc
  else (
    let obj_name = get_obj_name_from_aexpr obj in
    if obj_name = ""
    then acc
    else (
      let encoded_list_name = encode_list_name list_name in
      add_new_equality acc encoded_list_name obj_name
    )
  )

and obj_collect_equalities_dict (acc: ObjectTypeSet.t list) (dict: aexpr) (obj: aexpr) =
  let dict_name = get_dict_name_from_aexpr dict in
  if dict_name = ""
  then acc
  else (
    let obj_name = get_obj_name_from_aexpr obj in
    if obj_name = ""
    then acc
    else (
      let encoded_dict_name = encode_dict_name dict_name in
      add_new_equality acc encoded_dict_name obj_name
    )
  )

and obj_collect_equalities_aexpr (acc: ObjectTypeSet.t list) (aexpr: aexpr) : (ObjectTypeSet.t list) =
  match aexpr with
    (* Case 1; obj1 = obj2 -> (obj1, obj2) *)
    | AAssign(lhs_obj, aexpr, _) ->
      let acc = obj_collect_equalities_aexpr acc aexpr in
      (match aexpr with
        | AVal(rhs_obj, aval_data_type) ->
          (match aval_data_type with
            | Object(_) ->
              let acc = add_new_equality acc lhs_obj rhs_obj in
              acc
            | _ -> acc
          )
        | ACall(fname, _, _) ->
          insert_obj_into_fname_to_obj_equality_hash acc fname lhs_obj
        | AListDecl(ae_list, typ) ->
          (match typ with
            | List(Object(_)) ->
              List.fold_left (fun acc ae -> add_new_equality acc (encode_list_name lhs_obj) (get_obj_name_from_aexpr ae)) acc ae_list
            | _ -> acc
          )
        | AListRemove(lst, _, typ) ->
          (match typ with
            | Object(_) ->
              let acc = add_new_equality acc lhs_obj (encode_list_name (get_list_name_from_aexpr lst)) in
              acc
            | _ -> acc
          )
        | AListPop(lst, typ) ->
          (match typ with
            | Object(_) ->
              let acc = add_new_equality acc lhs_obj (encode_list_name (get_list_name_from_aexpr lst)) in
              acc
            | _ -> acc
          )
        | AListGet(lst, _, typ) ->
          (match typ with
            | Object(_) ->
              let acc = add_new_equality acc lhs_obj (encode_list_name (get_list_name_from_aexpr lst)) in
              acc
            | _ -> acc
          )
        | ADictDelete(dict, _, typ) ->
          (match typ with
            | Object(_) ->
              let acc = add_new_equality acc lhs_obj (encode_dict_name (get_dict_name_from_aexpr dict)) in
              acc
            | _ -> acc
          )
        | ADictFind(dict, _, typ) ->
          (match typ with
            | Object(_) ->
              let acc = add_new_equality acc lhs_obj (encode_dict_name (get_dict_name_from_aexpr dict)) in
              acc
            | _ -> acc
          )
        | AObjectField(_, field, typ) ->
          (match typ with
            | Object(_) ->
              let acc = add_new_equality acc lhs_obj field in
              acc
            | _ -> acc
          )
        | _ -> acc
      )
    (* Case 2; foo(obj1), def foo(a) -> (obj1, a) *)
    | ACall(func_name, args, _) ->
      let acc = obj_collect_equalities_params func_name args acc in
      if Hashtbl.mem fname_to_visited_hash func_name
      then acc
      else let _ = Hashtbl.add fname_to_visited_hash func_name true in
      (obj_collect (Hashtbl.find afunc_decl_hash func_name)) acc
    (* Case 3; if obj1 == obj2 -> (obj1, obj2) *)
    | ABinop(lhs, op, rhs, _) ->
      let acc = obj_collect_equalities_binop acc lhs op rhs in
      let acc = obj_collect_equalities_aexpr acc lhs in
      let acc = obj_collect_equalities_aexpr acc rhs in
      acc
    (* Case 4; a.obj1 = obj2 -> (obj1, obj2) *)
    | AObjectAssign(_, field, ae2, _) ->
      (match ae2 with
        | AVal(name, aval_typ) ->
          (match aval_typ with
            | Object(_) -> add_new_equality acc field name
            | _ -> acc
          )
        | _ -> acc
      )

    (* List Function Cases *)
    | AListInsert(lst, _, obj, _) ->
      obj_collect_equalities_list acc lst obj
    | AListPush(lst, obj, _) ->
      obj_collect_equalities_list acc lst obj
    | AListSet(lst, _, obj, _) ->
      obj_collect_equalities_list acc lst obj

    (* Dict Function Cases *)
    | ADictMap(dict, _, obj, _) ->
      obj_collect_equalities_dict acc dict obj

    (* Fallthrough *)
    | _ -> let _ = log ("[obj_collect_equalities_aexpr] Fallthrough with " ^ (string_of_aexpr aexpr)) in acc

(*
obj_collect_equalities_astmt: Collects the equality list from an astmt
*)
and obj_collect_equalities_astmt (acc: ObjectTypeSet.t list) (annotated_stmt: astmt) (fname: string) : (ObjectTypeSet.t list) =
  let _ = print_set_list acc in
  match annotated_stmt with
    | ABlock(astmt_list) ->
      List.fold_left (fun acc astmt -> obj_collect_equalities_astmt acc astmt fname) acc astmt_list
    | AExpr(ae) -> obj_collect_equalities_aexpr acc ae
    | AReturn(ae) ->
      let acc = match ae with
        | AVal(ret_name, typ) ->
          (match typ with
            | Object(_) ->
              let acc = insert_obj_into_fname_to_obj_equality_hash acc fname ret_name in
              acc
            | _ -> acc
          )
        | _ -> acc
      in
      obj_collect_equalities_aexpr acc ae
    | AWhile(ae, _) -> obj_collect_equalities_aexpr acc ae
    | AForIn(ae1, ae2, body) ->
      let new_acc = obj_collect_equalities_aexpr acc ae1 in
      let newer_acc = obj_collect_equalities_aexpr new_acc ae2 in
      let newest_acc = obj_collect_equalities_astmt newer_acc body fname in
      newest_acc
    | AForRange(ae1, ae2, ae3, body) ->
      let new_acc = obj_collect_equalities_aexpr acc ae1 in
      let newer_acc = obj_collect_equalities_aexpr new_acc ae2 in
      let newest_acc = obj_collect_equalities_aexpr newer_acc ae3 in
      let newester_acc = obj_collect_equalities_astmt newest_acc body fname in
      newester_acc
    | AIf(ae, astmt1, astmt2) ->
      let new_acc = obj_collect_equalities_aexpr acc ae in
      let newer_acc = obj_collect_equalities_astmt new_acc astmt1 fname in
      let newest_acc = obj_collect_equalities_astmt newer_acc astmt2 fname in
      newest_acc
    | AObjectInit(obj_list) -> obj_collect_fields_obj_list obj_list acc
    | _ -> acc

and obj_collect_fields_obj_list (obj_list: (string * data_type) list) (acc: ObjectTypeSet.t list) : (ObjectTypeSet.t list) =
  match obj_list with
    | [] -> acc
    | [obj] -> add_lone_equality acc (fst obj)
    | obj1 :: obj2 :: tail -> let new_acc = (add_new_equality acc (fst obj1) (fst obj2)) in
      (obj_collect_fields_obj_list (obj2 :: tail) new_acc)


(*
obj_collect_equalities: Collects a list of object type sets based on operations between objects
params:
  astmts: an astmt list to collect equalities from
returns:
  a list of sets where each set contains objects that are of the same type
*)
and obj_collect_equalities (astmts: astmt list) (acc: ObjectTypeSet.t list) (fname: string) : (ObjectTypeSet.t list) =
  let acc = List.fold_left (fun acc astmt -> obj_collect_equalities_astmt acc astmt fname) acc astmts  in
  acc

(*
[1]

obj_collect: Populate obj_to_fields_hash and obj_equality_sets
params:
  afdecl: a function declaration to collect from
  obj_equality_sets: the list of sets to populate
returns:
  the populated obj_equality_sets
*)
and obj_collect (afdecl: Sast.afunc_decl) (acc: ObjectTypeSet.t list) : (ObjectTypeSet.t list) =
  let _ = List.map obj_collect_fields_astmt afdecl.abody in
  let obj_equality_sets = obj_collect_equalities afdecl.abody acc afdecl.afname in
  obj_equality_sets






(* ==================== OBJECT UNIFY ==================== *)

(*
obj_unify_one_set: Populate obj_to_id_hash and id_to_fields_hash for one set
params:
  set: a set of objects of the same type
  id: the current available type id
returns:
  unit
*)
and obj_unify_one_set (set: ObjectTypeSet.t) (id: int) =
  (* Populate id_to_fields hash *)
  let get_obj_fields obj =
    if Hashtbl.mem obj_to_fields_hash obj then Hashtbl.find obj_to_fields_hash obj else []
  in
  let fields = List.fold_left (fun acc obj -> acc @ (get_obj_fields obj)) [] (ObjectTypeSet.elements set) in
  let _ = Hashtbl.add id_to_fields_hash id fields in

  (* Populate obj_to_id_hash *)
  ObjectTypeSet.iter
    (fun x -> let _ = log ("[obj_unify_one_set] adding " ^ x ^ "->" ^ (string_of_int id)) in
    Hashtbl.add obj_to_id_hash x id)
  set

and obj_unify_lone_object (obj: string) (fields: string list) =
  if Hashtbl.mem obj_to_id_hash obj
  then ()
  else
    let obj_id = gen_new_obj_id() in
    let _ = Hashtbl.add obj_to_id_hash obj obj_id in
    let _ = Hashtbl.add id_to_fields_hash obj_id fields in
    ()

(*
[2]

obj_unify: Populate obj_to_id_hash and id_to_fields_hash for all sets
params:
  obj_equality_sets: the list of sets to populate
returns:
  unit
*)
and obj_unify (obj_equality_sets: ObjectTypeSet.t list) =
  let _ = List.iter (fun x -> let obj_id = gen_new_obj_id() in obj_unify_one_set x obj_id) obj_equality_sets in
  Hashtbl.iter (fun k v -> obj_unify_lone_object k v) obj_to_fields_hash






(* ==================== OBJECT APPLY ==================== *)

and obj_apply_aexpr (aexpr: Sast.aexpr) : (Sast. aexpr) =
  match aexpr with
    | AVal(name, typ) ->
      if Hashtbl.mem obj_to_id_hash name
      then AVal(name, Object(Hashtbl.find obj_to_id_hash name))
      else (
        let encoded_list_name = encode_list_name name in
        if Hashtbl.mem obj_to_id_hash encoded_list_name
        then AVal(name, List(Object(Hashtbl.find obj_to_id_hash encoded_list_name)))
        else AVal(name, typ)
      )
    | ABinop(lhs, op, rhs, typ) ->
      let new_lhs = obj_apply_aexpr lhs in
      let new_rhs = obj_apply_aexpr rhs in
      ABinop(new_lhs, op, new_rhs, typ)
    | AUnop(op, ae, _) ->
      let new_ae = obj_apply_aexpr ae in
      AUnop(op, new_ae, type_of_aexpr new_ae)
    | AAssign(id, rhs, _) ->
      let new_rhs = obj_apply_aexpr rhs in
      if is_object_list_decl rhs = true && Hashtbl.mem obj_to_id_hash (encode_list_name id) = true
      then
        let new_typ = List(Object(Hashtbl.find obj_to_id_hash (encode_list_name id))) in
        let new_list_decl = inject_list_decl_type new_rhs new_typ in
        AAssign(id, new_list_decl, new_typ)
      else AAssign(id, new_rhs, type_of_aexpr new_rhs)
    | ACall(fname, args, typ) ->
      (match typ with
        | Object(_) -> ACall(fname, (List.map obj_apply_aexpr args), Hashtbl.find func_type_hash fname)
        | _ -> ACall(fname, (List.map obj_apply_aexpr args), typ)
      )
    (* Lists *)
    | AListDecl(ae_list, typ) ->
      let new_ae_list = List.map obj_apply_aexpr ae_list in
      AListDecl(new_ae_list, get_list_type_from_listdecl (AListDecl(new_ae_list, typ)))
    | AListInsert(ae1, ae2, ae3, _) ->
      let new_ae1 = obj_apply_aexpr ae1 in
      let new_ae2 = obj_apply_aexpr ae2 in
      let new_ae3 = obj_apply_aexpr ae3 in
      AListInsert(new_ae1, new_ae2, new_ae3, type_of_aexpr new_ae3)
    | AListPush(ae1, ae2, _) ->
      let new_ae1 = obj_apply_aexpr ae1 in
      let new_ae2 = obj_apply_aexpr ae2 in
      AListPush(new_ae1, new_ae2, type_of_aexpr new_ae2)
    | AListRemove(ae1, ae2, _) ->
      let new_ae1 = obj_apply_aexpr ae1 in
      let new_ae2 = obj_apply_aexpr ae2 in
      AListRemove(new_ae1, new_ae2, type_of_aexpr new_ae2)
    | AListPop(ae, typ) ->
      let new_ae = obj_apply_aexpr ae in
      let obj_id = get_list_type_from_aexpr new_ae in
      if obj_id = -1
      then AListPop(new_ae, typ)
      else AListPop(new_ae, Object(obj_id))
    | AListDequeue(ae, typ) ->
      let new_ae = obj_apply_aexpr ae in
      let obj_id = get_list_type_from_aexpr new_ae in
      if obj_id = -1
      then AListDequeue(new_ae, typ)
      else AListDequeue(new_ae, Object(obj_id))
    | AListLength(ae, typ) ->
      let new_ae = obj_apply_aexpr ae in
      AListLength(new_ae, typ)
    | AListGet(ae1, ae2, typ) ->
      let new_ae1 = obj_apply_aexpr ae1 in
      let new_ae2 = obj_apply_aexpr ae2 in
      let obj_id = get_list_type_from_aexpr new_ae1 in
      if obj_id = -1
      then AListGet(new_ae1, new_ae2, typ)
      else AListGet(new_ae1, new_ae2, Object(obj_id))
    | AListSet(ae1, ae2, ae3, _) ->
      let new_ae1 = obj_apply_aexpr ae1 in
      let new_ae2 = obj_apply_aexpr ae2 in
      let new_ae3 = obj_apply_aexpr ae3 in
      AListSet(new_ae1, new_ae2, new_ae3, type_of_aexpr new_ae3)

    (* Dicts *)
    (* TODO ADictDecl *)
    | ADictMem(ae1, ae2, typ) ->
      let new_ae1 = obj_apply_aexpr ae1 in
      let new_ae2 = obj_apply_aexpr ae2 in
      ADictMem(new_ae1, new_ae2, typ)
    | ADictDelete(ae1, ae2, typ) ->
      let new_ae1 = obj_apply_aexpr ae1 in
      let new_ae2 = obj_apply_aexpr ae2 in
      let obj_id = get_dict_val_type_from_aexpr new_ae1 in
      if obj_id = -1
      then ADictDelete(new_ae1, new_ae2, typ)
      else ADictDelete(new_ae1, new_ae2, Object(obj_id))
    | ADictSize(ae, typ) ->
      let new_ae = obj_apply_aexpr ae in
      ADictSize(new_ae, typ)
    | ADictFind(ae1, ae2, typ) ->
      let new_ae1 = obj_apply_aexpr ae1 in
      let new_ae2 = obj_apply_aexpr ae2 in
      let obj_id = get_dict_val_type_from_aexpr new_ae1 in
      if obj_id = -1
      then ADictFind(new_ae1, new_ae2, typ)
      else ADictFind(new_ae1, new_ae2, Object(obj_id))
    | ADictMap(ae1, ae2, ae3, typ) ->
      let new_ae1 = obj_apply_aexpr ae1 in
      let new_ae2 = obj_apply_aexpr ae2 in
      let new_ae3 = obj_apply_aexpr ae3 in
      ADictMap(new_ae1, new_ae2, new_ae3, typ)

    (* Objects *)
    | AObjectField(ae, field, typ) ->
      let new_ae = obj_apply_aexpr ae in
      let new_typ =
        if Hashtbl.mem obj_to_id_hash field
        then Object(Hashtbl.find obj_to_id_hash field)
        else typ
      in
      AObjectField(new_ae, field, new_typ)
    | AObjectAssign(lhs_aexpr, field, rhs_aexpr, typ) ->
      let lhs = obj_apply_aexpr lhs_aexpr in
      let rhs = obj_apply_aexpr rhs_aexpr in
      let new_typ = match typ with
        | Object(_) -> let _ = Hashtbl.add field_hash field (type_of rhs) in type_of rhs
        | List(t) -> (match t with
          | Object(_) -> type_of rhs
          | _ -> typ)
        | Dict(_, t) -> (match t with
          | Object(_) -> type_of rhs
          | _ -> typ)
        | _ -> typ
      in AObjectAssign(lhs, field, rhs, new_typ)
    | _ -> aexpr

(*
obj_apply_stmt: Augment one statement with object types
params:
  astmt: the astmt to augment
  afname: the name of the enclosing function
returns:
  the augmented astmt
*)
and obj_apply_astmt (astmt: Sast.astmt) (afname: string) : (Sast.astmt) =
  match astmt with
    | ABlock(astmt_list) -> ABlock(List.map (fun astmt -> obj_apply_astmt astmt afname) astmt_list)
    | AExpr(ae) -> AExpr(obj_apply_aexpr ae)
    | AReturn(ae) ->
      let new_ae = obj_apply_aexpr ae in
      let typ = type_of_aexpr new_ae in
      let _ =
        (match typ with
          | Object(id) ->  let _ = log ("Labeling the return type of " ^ afname ^ " to be " ^ string_of_int(id)) in Hashtbl.add func_type_hash afname (Object(id))
          | _ -> ()
        )
      in AReturn(new_ae)
    | AWhile(ae, astmt) -> AWhile(obj_apply_aexpr ae, obj_apply_astmt astmt afname)
    | AForIn(ae1, ae2, astmt) -> AForIn(obj_apply_aexpr ae1, obj_apply_aexpr ae2, obj_apply_astmt astmt afname)
    | AForRange(ae1, ae2, ae3, astmt) -> 
        AForRange(obj_apply_aexpr ae1, obj_apply_aexpr ae2, obj_apply_aexpr ae3, obj_apply_astmt astmt afname)
    | AIf(ae, astmt1, astmt2) -> AIf(obj_apply_aexpr ae, obj_apply_astmt astmt1 afname, obj_apply_astmt astmt2 afname)
    | APrint(ae) -> APrint(obj_apply_aexpr ae)
    | AObjectInit(obj_list) -> AObjectInit(obj_apply_obj_list obj_list)
    | _ -> astmt

and obj_apply_obj_list (obj_list: (string * data_type) list) =
  List.map (fun obj -> 
    match obj with
      | (name, Object(_)) ->
        if Hashtbl.mem obj_to_id_hash (fst obj)
        then (name, Object(Hashtbl.find obj_to_id_hash (fst obj)))
        else raise (failwith("Object has no id in obj_apply_obj_list"))
      | _ -> raise (failwith("Attempting to ObjectInit a non-Object"))
  ) obj_list

and obj_apply_aformals (aformal: (data_type * string)) : (data_type * string) =
  let typ = fst aformal in
  let name = snd aformal in
  if Hashtbl.mem obj_to_id_hash name
  then (Object(Hashtbl.find obj_to_id_hash name), name)
  else (
    let list_name = encode_list_name name in
    if Hashtbl.mem obj_to_id_hash list_name
    then (List(Object(Hashtbl.find obj_to_id_hash list_name)), name)
    else (typ, name)
  )

(*
obj_apply_function: Augment one afdecl with object types
params:
  afdecl: the afdecl to augment
returns:
  the augmented afdecl
*)
and obj_apply_function (afdecl: Sast.afunc_decl) : (Sast.afunc_decl) =
  let new_afunc_decl = {
    afname = afdecl.afname;
    aformals = List.map obj_apply_aformals afdecl.aformals;
    abody = List.map (fun astmt -> obj_apply_astmt astmt afdecl.afname) afdecl.abody;
    return = afdecl.return;
  } in new_afunc_decl

(* Second pass to apply return types *)
and obj_apply_return_types (afdecl: Sast.afunc_decl) : (Sast.afunc_decl) =
  let new_afunc_decl = {
    afname = afdecl.afname;
    aformals = afdecl.aformals;
    abody = List.map (fun astmt -> obj_apply_astmt astmt afdecl.afname) afdecl.abody;
    return =
      if Hashtbl.mem fname_to_obj_equality_hash afdecl.afname
      then
        let obj_list = Hashtbl.find fname_to_obj_equality_hash afdecl.afname in
        let arbitrary_obj = List.hd obj_list in
        if Hashtbl.mem obj_to_id_hash arbitrary_obj
        then Object(Hashtbl.find obj_to_id_hash arbitrary_obj)
        else afdecl.return
      else afdecl.return;
  } in new_afunc_decl

(*
[3]

obj_apply: Augment the Sast with type ids on all object nodes
params:
  sast: the list of afdecls to augment
returns:
  the augmented list of afdecls
*)
and obj_apply (sast: Sast.aprogram) =
  let first_sweep = List.map obj_apply_function sast in
  List.map obj_apply_return_types first_sweep
;;
