open Ast
open Printf

(* Unary operators *)
let txt_of_unop = function
    | Neg -> "Neg"
    | Not -> "Not"

(* Binary operators *)
let txt_of_binop = function
  (* Arithmetic *)
  | Add -> "Add"
  | Minus -> "Minus"
  | Times-> "Times"
  | Divide -> "Divide"
  (* Boolean *)
  | Or -> "Or"
  | And -> "And"
  | Eq -> "Eq"
  | Neq -> "Neq"
  | Less -> "Less"
  | Leq -> "Leq"
  | Greater -> "Greater"
  | Geq -> "Geq"
  (* List *)
  | Combine  -> "Combine"
  (* String *)
  | Concat -> "Concat"

(* Expressions *)
(*let txt_of_num = function
  | Num_int(x) -> string_of_int x
  | Num_float(x) -> string_of_float x*)

let rec txt_of_expr = function
  | Num_lit(x) -> sprintf "Num_lit(%s)" (string_of_float x)
  | String_lit(x) -> sprintf "String_lit(%s)" x
  | Bool_lit(x) -> sprintf "Bool_lit(%s)" (string_of_bool x)
  (* | None_lit -> sprintf "None lit" *)
  | Id(x) -> sprintf "Id(%s)" x
  | Unop(op, e) -> sprintf "Unop(%s, %s)" (txt_of_unop op) (txt_of_expr e)
  | Binop(e1, op, e2) -> sprintf "Binop(%s, %s, %s)"
      (txt_of_expr e1) (txt_of_binop op) (txt_of_expr e2)
  | Call(str, args) -> sprintf "Call(%s, [%s])"
      str (txt_of_list args)
  | Assign(x, e) -> sprintf "Assign(%s, %s)" x (txt_of_expr e)
  | ListDecl(l) -> sprintf "ListDecl([%s])" (txt_of_list l)
  (* List operations *)
  | ListInsert(e1, e2, e3) -> sprintf "ListInsert(%s, %s, %s)" (txt_of_expr e1) (txt_of_expr e2) (txt_of_expr e3)
  | ListPush(e1, e2) -> sprintf "ListPush(%s, %s)" (txt_of_expr e1) (txt_of_expr e2)
  | ListRemove(e1, e2) -> sprintf "ListRemove(%s, %s)" (txt_of_expr e1) (txt_of_expr e2)
  | ListPop(e) -> sprintf "ListPop(%s)" (txt_of_expr e)
  | ListDequeue(e) -> sprintf "ListDequeue(%s)" (txt_of_expr e)
  | ListLength(e) -> sprintf "ListLength(%s)" (txt_of_expr e)
  | ListGet(e1, e2) -> sprintf "ListGet(%s, %s)" (txt_of_expr e1) (txt_of_expr e2)
  | ListSet(e1, e2, e3) -> sprintf "ListSet(%s, %s, %s)" (txt_of_expr e1) (txt_of_expr e2) (txt_of_expr e3)
  (* Dict Operations *)
  | DictDecl(d) -> sprintf "DictDecl([%s])" (txt_of_dict d)
  | DictMem(e1, e2) -> sprintf "DictMem(%s, %s)" (txt_of_expr e1) (txt_of_expr e2)
  | DictFind(e1, e2) -> sprintf "DictFind(%s, %s)" (txt_of_expr e1) (txt_of_expr e2)
  | DictMap(e1, e2, e3) -> sprintf "DictMap(%s, %s, %s)" (txt_of_expr e1) (txt_of_expr
  e2) (txt_of_expr e3)
  | DictDelete(e1, e2) -> sprintf "DictDelete(%s, %s)" (txt_of_expr e1) (txt_of_expr e2)
  | DictSize(e) -> sprintf "DictSize(%s)" (txt_of_expr e)
  (*| Assert(e) -> sprintf "Assert([%s])" (txt_of_expr e) *)
  | Noexpr -> sprintf "Noexpr"
  | ObjectField(e, s2) -> sprintf "ObjectField(%s, %s)" (txt_of_expr e) s2
  | ObjectAssign(e1, s2, e2) -> sprintf "ObjectAssign(%s, %s, %s)" (txt_of_expr e1) s2 (txt_of_expr e2)


(* Lists *)
and txt_of_list = function
  | [] -> ""
  | [x] -> txt_of_expr x
  | _ as l -> String.concat " ; " (List.map txt_of_expr l)

and txt_of_tup (tup: expr * expr) =
  sprintf "(%s, %s)" (txt_of_expr (fst tup)) (txt_of_expr (snd tup))

and txt_of_dict = function
  | [] -> ""
  | [t] -> (txt_of_tup t)
  | _ as d -> String.concat " ; " (List.map txt_of_tup d)

and txt_of_stringlist (stringlist: string list) =
  let rec aux acc = function
    | [] -> sprintf "%s" (String.concat ", " (acc))
    | hd :: tl -> aux (hd :: acc) tl
  in aux [] stringlist

(* Statements *)
and txt_of_stmt (st: stmt) =
  match st with
    | Expr(e) -> sprintf "%s" (txt_of_expr e)
    | Return(e) -> sprintf "Return(%s)" (txt_of_expr e)
    | Break -> sprintf "Break"
    | Continue -> sprintf "Continue"
    | While(e, s) -> sprintf "While(%s, %s)" (txt_of_expr e) (txt_of_stmt s)
    | ForIn(e1, e2, s) -> sprintf "ForIn(%s, %s, %s)" (txt_of_expr e1) (txt_of_expr e2) (txt_of_stmt s)
    | ForRange(e1, e2, e3, s) ->
        sprintf "ForRange(%s, %s, %s, %s)" (txt_of_expr e1) (txt_of_expr e2)
                                           (txt_of_expr e3) (txt_of_stmt s)
    | If(e1, s1, s2) -> sprintf "If(%s, %s, %s)"
      (txt_of_expr e1) (txt_of_stmt s1) (txt_of_stmt s2)
    | Block(sl) -> (txt_of_stmts sl)
    | Print(e) -> sprintf "Print([%s])" (txt_of_expr e)
    | ObjectInit(obj_list) -> sprintf "ObjectInit([%s])" (txt_of_stringlist obj_list)

and txt_of_stmts (stmts: stmt list): string =
  let rec aux acc = function
      | [] -> sprintf "[%s]" (String.concat " ; " (List.rev acc))
      | stmt :: tl -> aux (txt_of_stmt stmt :: acc) tl
  in aux [] stmts

(* Function declarations *)
and txt_of_fdecl (f: func_decl): string =
    sprintf "Fdecl({ params=[%s] ; body=%s})" (String.concat " ; " f.formals) (txt_of_stmts f.body)


and txt_of_fdecls fdecls =
  let rec aux acc = function
      | [] -> sprintf "[%s]" (String.concat " ; " (List.rev acc))
      | fdecl :: tl -> aux ((txt_of_fdecl fdecl) :: acc) tl
  in aux [] fdecls


(* Program entry point *)
let _ =
  let lexbuf = Lexing.from_channel stdin in
  let program = Parser.program Scanner.token lexbuf in
  let result = txt_of_fdecls program in
  print_endline result
